/*
 * $Id: arch-solaris.h,v 1.2 1997/03/01 17:42:35 athena Exp $
 */
/*
 * $Log: arch-solaris.h,v $
 * Revision 1.2  1997/03/01 17:42:35  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef SPARC_V9
#define SPARC_V9 0
#endif

#ifndef SPARC_V8
#define SPARC_V8 (!(SPARC_V9))
#endif

#define CACHE_LINE 64

/* padding, to be used in structures to avoid false sharing */
#define CILK_CACHE_LINE_PAD  char __dummy[CACHE_LINE]

#define CACHE_LINE_ALIGNED(decl) decl  __attribute__((aligned(CACHE_LINE)))

/* safety test */
#if SPARC_V8 && SPARC_V9
#error define either SPARC_V8 or SPARC_V9, but not both
#endif

#if (!SPARC_V8) && (!SPARC_V9)
#error define either SPARC_V8 or SPARC_V9
#endif

#if (!CILK_SOLARIS)
#error CILK_SOLARIS not defined !
#endif

/***********************************************************\
 * Lock implementation
\***********************************************************/
#include <thread.h>
#include <synch.h>

/* a lock variable */
typedef union {
     volatile int memory;
     mutex_t os;
} Cilk_mutex;

/***********************************************************\
 * Machine size
\***********************************************************/

#include <unistd.h>

static inline int Cilk_partition_size(void)
{
     /* 
      * This is only really right for an unloaded machine.
      */
     return sysconf(_SC_NPROCESSORS_ONLN);
}

/*
 * Get OS page size.
 */
static inline size_t Cilk_get_page_size(void)
{
     return sysconf(_SC_PAGESIZE);
}

/***********************************************************
 * Interaction with unix
 ***********************************************************/
#define Cilk_yield()     { yield(); }

/***********************************************************\
 * Various types of membars
\***********************************************************/

#if SPARC_V8
/* helper function - a hardware swap instruction */
static inline int Cilk_swap(register volatile int *ptr, register int xchg)
{
     asm volatile ("swap [%1],%0"
		   :"=r" (xchg)	/* registers written */
		   :"r"(ptr), "0"(xchg)		/* registers read */
     );
     return xchg;
}
#endif

#if SPARC_V9
/*
 * Use of SWAP is deprecated on V9 (see manual, A.56).
 * The manual suggests to implement locks with CAS, rather than
 * using SWAP.
 *
 * CAS (Compare And Swap) works as follows: CAS r0, r1, r2
 * reads [r0], compares with r1, if they are the same swaps
 * [r0] and r2, otherwise loads r2 with [r0] and leaves [r0]
 * unchanged.  Everything happens atomically, of course.
 * Also, according to the manual (8.4.4.3), "atomic load-stores
 * are ordered with respect to load and stores", which should
 * ensure the release consistency we want.
 *
 * Experimentally, CAS seems to be a bit faster than SWAP.
 * I don't know if this happens because CAS saves the extra store
 * (sometimes).
 */
static inline int Cilk_cas(register volatile int *ptr, register int cmp,
			   register int xchg)
{
     asm volatile ("cas [%1],%3,%0"
		   :"=r" (xchg)	/* registers written */
		   :"r"(ptr), "0"(xchg), "r"(cmp) /* registers read */
     );
     return xchg;
}
#endif

/*
 *  Ensure that all previous memory operations are completed before
 *  continuing.
 */
static inline void Cilk_fence(void)
{
#if SPARC_V9
     asm volatile ("membar #StoreLoad | #StoreStore | #LoadLoad | #LoadStore"
		   ::);
#endif
#if SPARC_V8
     /* 
      * this should be ok, maybe there is a faster way.
      * Notice that we are using SWAP only for its side effect:
      * atomic operations are ordered with respect to loads and
      * stores.
      */
     int dummy;
     Cilk_swap(&dummy, 0);
#endif
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future writes become visible.
 */
static inline void Cilk_membar_StoreStore(void)
{
     /* don't need to do anything for TSO (Total Store Order) */
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_StoreLoad(void)
{
#if SPARC_V9
     asm volatile ("membar #StoreLoad"::);
#endif
#if SPARC_V8
     /* 
      * The only way I know to implement the membar on a V8 sparc
      * is to use swap. Maybe there is a better way... --- athena
      */
     Cilk_fence();
#endif
}

/***********************************************************\
 * Timing functions
\***********************************************************/

#include <sys/time.h>

/* type to hold a high-resolution timer.  Must be an arithmetic type. */
typedef hrtime_t Cilk_time;

static inline Cilk_time Cilk_get_time(void)
{
     return gethrvtime();
}

static inline double Cilk_time_to_sec(Cilk_time t)
{
     return (double) t / 1.0e9;
}

static inline Cilk_time Cilk_get_wall_time(void)
{
     return gethrtime();
}

static inline double Cilk_wall_time_to_sec(Cilk_time t)
{
     return (double) t / 1.0e9;
}


